#include <eigen3/Eigen/Dense>
#include <eigen3/Eigen/Sparse>
#include <eigen3/Eigen/SparseLU>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <functional>
#include <vector>
#include <chrono>
using namespace std::chrono;
using namespace Eigen;

//! Solving the possibly overdetermined linear system of equations
//! Ax = b by solving normal equations
//! The least squares solution is returned
////TASK A
VectorXd normalEquationSolve(const MatrixXd &A, const VectorXd &b){
    if (b. size () != A.rows() ) throw std::runtime_error("Dimension mismatch");  // Cholesky solver
    VectorXd x = (A.transpose()*A).llt().solve(A.transpose()*b);
    return x ;
}
////END OF TASK A

////TASK C
//! Given a sparse matrix A, construct the sparse matrix C for the modified system Cy=d
SparseMatrix<double> computeC(SparseMatrix<double> &A, const VectorXd &b){
    int m = A.rows();
    int n = A.cols();
    int sizeC = m+n;
    std :: vector < Triplet <double > > triplets_C ;
    triplets_C.reserve(m+2*A.nonZeros()) ;
    
    // Fill the triplets with elements from A and A_transpose
    for(int i = 0; i < A.outerSize(); ++i){
        for(SparseMatrix<double>::InnerIterator it(A,i); it; ++it){
            triplets_C.emplace_back(it.row(),it.col()+m,it.value());
            triplets_C.emplace_back(it.col()+m,it.row(),it.value());
        }
    }
    
    //Fill the triplets corresponding to the Identity Block
    for (int k=0; k<m; ++k){
        triplets_C.push_back(Triplet<double>(k ,k ,-1.0));
    }
    
    SparseMatrix <double> C(sizeC, sizeC) ;
    C.setFromTriplets(triplets_C.begin(), triplets_C.end());
    C.makeCompressed();
    return C;
}

//! Given the original RHS-vector b, construct the RHS-vector d for the modified system Cy=d
VectorXd computeRHS(SparseMatrix<double> &A, const VectorXd &b){
    int m = A.rows();
    int n = A.cols();
    int sizeC = m+n;
    VectorXd d = VectorXd :: Zero(sizeC);
    d.head(m) = b;
    
    return d ;
}
////END OF TASK C

////TASK D
//Obtain solution for the modified system using a SparseLU solver
VectorXd computeSolution(SparseMatrix<double> &A, const VectorXd &b){
    VectorXd x, y;
    VectorXd d = computeRHS(A, b);
    SparseMatrix<double> C = computeC(A, b);
    // Perform sparse elimination
    SparseLU<SparseMatrix<double>> solver(C);
    y = solver.solve(d);
    
    return x = y.tail(A.cols()); //The actual solution vector 'x' is the last 'n' elements of y
}
////END OF TASK D

double f(double x) {
    return M_PI*M_PI*std::sin(M_PI*x); //returns value of RHS-function f at x
}

////TASK F
//Construct dense matrix 'A' obtained through Finite Difference Method for the poisson problem
MatrixXd constructDenseA_Poisson(int N){
    // Mesh width
    double delta_x = 1.0/(N+1);
    MatrixXd A_dense = MatrixXd::Zero(N,N);
    
    A_dense.diagonal() = VectorXd::Constant(N,2);
    A_dense.diagonal<1>() = VectorXd::Constant(N-1,-1);
    A_dense.diagonal<-1>() = VectorXd::Constant(N-1,-1);
    A_dense *= 1/(delta_x*delta_x);
    
    return A_dense;
}

//Construct sparse matrix 'A' obtained through Finite Difference Method for the poisson problem
SparseMatrix <double> constructSparseA_Poisson(MatrixXd &A_dense, int N){
    SparseMatrix <double> A_sparse = A_dense.sparseView(); //Convert dense matrix A into a sparse form
    return A_sparse;
}
////END OF TASK F

int main() {
    // Mesh size (number of interior points)
    std::vector<int> N_interior = {25, 50, 75, 100, 200, 400, 800, 1600};
    for(unsigned int p = 0; p < N_interior.size(); ++p) {
        int N  = N_interior.at(p);
        double delta_x = 1.0/(N+1);

        VectorXd x(N+2), b(N), solution_exact(N);
        
        for(unsigned int i = 0; i < N + 2; ++i) {
            x(i) = i * delta_x; //Construct mesh vector
            
            if (i > 0 && i < N+1) {
                solution_exact(i-1) = std::sin(M_PI*x(i)); //Construct solution vector
                b(i-1) = f(x(i)); //Construct RHS vector 'b'
            }
        }
        
        
    //Obtain solution by standard normal equation and measure run-time
        MatrixXd A_dense = constructDenseA_Poisson(N);
        auto tic_normal =  high_resolution_clock::now();
        VectorXd solution_normal = normalEquationSolve(A_dense, b);
        auto toc_normal = high_resolution_clock::now();
        duration<double> runTime_normal = duration_cast<duration<double>>(toc_normal-tic_normal);
        
        
    //Obtain solution by modified system which preserves sparsity and measure run-time
        SparseMatrix <double> A_sparse = constructSparseA_Poisson(A_dense, N);
        auto tic_sparse =  high_resolution_clock::now();
        VectorXd solution_sparse = computeSolution(A_sparse, b);
        auto toc_sparse = high_resolution_clock::now();
        duration<double> runTime_sparse = duration_cast<duration<double>>(toc_sparse-tic_sparse);
        
    // Compute error in solution for both methods
        double maxerr_normal = (solution_exact - solution_normal).template lpNorm<Infinity>();
        double maxerr_sparse = (solution_exact - solution_sparse).template lpNorm<Infinity>();
        
    // I/O
        if (p == 0) std::cout << std::setw(15) << "N" << std::setw(15) << "maxerr_normal" << std::setw(15) << "runTime_normal" << std::setw(15) << "maxerr_sparse" << std::setw(15) << "runTime_sparse" << std::endl;
        std::cout << std::setw(15) << N << std::setw(15) << maxerr_normal << std::setw(15) << runTime_normal.count() << std::setw(15) << maxerr_sparse << std::setw(15) << runTime_sparse.count() << std::endl;
    }
}


// End of file
