#include <iostream>
#include <eigen3/Eigen/Dense>
//#include "gaussquad.hpp"


// test function
auto f_test = [](const double x)
                {
                    return (1.0 / (1.0 + 25.0*x*x));
                };
                
auto fm_test = [](const Eigen::MatrixXd x)
                {
                    return (1.0 / (1.0 + 25.0 * x.array() * x.array())).matrix();
                };


// Generate equidistant knots in the reference interval [0,1]
Eigen::VectorXd genKnotsEqui(const int n)
{
    return Eigen::VectorXd::LinSpaced(n, 0, 1);
}


// Generate Chebyshev knots in the reference interval [0,1]
Eigen::VectorXd genKnotsCheb(const int n)
{
    Eigen::VectorXd knots(n);
    for (int k=0; k<n; k++) {
        double theta = (2*(n-1-k)+1)/(2.*n)*M_PI;
        knots(k) = 0.5*(1 + cos(theta));
    }
    return knots;
}


// Piecewise Lagrange interpolation using Newton basis
class Newton {

public:
	
	Newton(const Eigen::VectorXd &grid_, const Eigen::MatrixXd &x) 
	    : grid(grid_), nodes(x)
	    { alpha.resize(x.rows(), x.cols()); }
	
	// computes Linf error
	// with respect to the given lambda function F
	template <typename LambdaFn>
	double EvalError(LambdaFn F)
    {
        double error = 0.0;
        
        const int n = nodes.rows();
        const int M = nodes.cols();
        const int nq = 11;
        Eigen::VectorXd ref_points = Eigen::VectorXd::LinSpaced(nq, -0.95, 0.95);
        
        for (int l=0; l<M; l++) {
        
            double a = grid(l);
            double b = grid(l+1);
            Eigen::VectorXd points = 0.5*(a + b) + 0.5*ref_points.array()*(b-a);
            
            for (int i=0; i<nq; i++) {
                error = fmax(fabs(F(points(i)) - this->Evaluate(l, points(i))), error);
                //std::cout << l << "\t" << i << "\t" 
                //          << fabs(F(points(i)) - this->Evaluate(l, points(i))) << "\t"
                //          << F(points(i)) << "\t" << this->Evaluate(l, points(i))<< std::endl;            
            }
            //std::cout << l << "\t" << error << std::endl;
        }
    
        return error;
    }
    
    /*
    // computes Linf error
	// with respect to the given lambda function F
    template <typename LambdaFn>
	double EvalError(LambdaFn F)
    {
        double error = 0.0;
        
        const int n = nodes.rows();
        const int M = nodes.cols();
        const int nq = ceil(0.5*n);
        QuadRule gauss = gaussquad_(nq);
        
        for (int l=0; l<M; l++) {
        
            double a = grid(l);
            double b = grid(l+1);
            Eigen::VectorXd points = 0.5*(a + b) + 0.5*gauss.nodes.array()*(b-a);
            Eigen::VectorXd weights = 0.5*(b-a)*gauss.weights; 
            //std::cout << "\n\n" << l << "\t" << a << "\t" << b << std::endl;
            //std::cout << points.transpose() << std::endl;
            //std::cout << weights.transpose() << std::endl;
            
            for (int i=0; i<nq; i++) {
                error += weights(i)*fabs(F(points(i)) - this->Evaluate(l, points(i)));
                //std::cout << l << "\t" << i << "\t" 
                //           << F(points(i)) << "\t" << this->Evaluate(l, points(i))<< std::endl;            
            }
        }
    
        return error;
    }*/
	
	// returns the Lagrange coefficients
	Eigen::MatrixXd GetCoeffs()
	{
	    return this->alpha;
	}
	
	
	void Interpolate(const Eigen::MatrixXd &y);
	double Evaluate(const int l, double x);

private:
    
    Eigen::VectorXd grid;
    Eigen::MatrixXd nodes; // nodes
	Eigen::MatrixXd alpha;	// coefficients
};


////TASK C
// Generate nodes in interval [a,b]
// Input:
/// M     : number of sub-intervals.
/// knots : equidistant knots in the reference interval [0,1].
// Output: nodes
Eigen::MatrixXd genNodes(const double a, const double b, const int M, const Eigen::VectorXd &knots)
{
    const int n = knots.size();
    double H = (b-a)/M;
    
    Eigen::MatrixXd nodes(n,M);
    for (int j=0; j<M; j++) {
        nodes.col(j) = (a + j*H) + knots.array()*H;
    }
    
    return nodes;
}
////END OF TASK C


////TASK D
////TASK D_1
// Computes the coefficients `alpha' in the Newton basis
// for piecewise Lagrange interpolation
// Input:
/// y     : function values at the nodes
void Newton::Interpolate(const Eigen::MatrixXd &y)
{
    int n = nodes.rows();
    int M = nodes.cols();
    assert(y.rows() == n);
    assert(y.cols() == M);
    
    // set-up and solve linear system
    for (int l=0; l<M; l++)
    {
        Eigen::MatrixXd A(n,n);
        A.setZero();
        
        for (int i=0; i<n; i++) {
            A(i,0) = 1;
            for (int j=1; j<=i; j++) {
                A(i,j) = A(i,j-1)*(nodes(i,l) - nodes(j-1,l));
            }
        }
        
        alpha.col(l) = A.triangularView<Eigen::Lower>().solve(y.col(l));
    }
    
    // solve with divided difference method
    /*for (int l=0; l<M; l++) {
    
        alpha.col(l) = y.col(l);
	    for (int j=0; j<n-1; ++j) {
		    for (int i=n-1; i>j; --i) {
			    alpha(i,l) = (alpha(i,l) - alpha(i-1,l)) / (nodes(i,l) - nodes(i-1-j,l));
		    }
	    }
    }*/
}
////END OF TASK D_1

////TASK D_2
// Evaluates the piecewise Lagrange interpolant
// at a given point 'x' in the element 'l'
// Input:
/// l     : element index, this element contains the point x
/// x     : given point
double Newton::Evaluate(const int l, double x)
{
	int n = alpha.rows();
	double y = alpha(n-1,l);
	for (int i = n-2; i >= 0; --i) { // Horner scheme
		y = y * (x - nodes(i,l)) + alpha(i,l);
	}
	
	return y;
}
////END OF TASK D_2
////END OF TASK D

// Given function f
auto f = [](const double x)
           {
                if (x < 1./3.)
                    return (1.0 / (1.0 + 25.0 * (6*x-1) * (6*x-1)));
                else
                    return (1.0 + cos(M_PI*x));
           };

// function f1
auto f1m = [](const Eigen::MatrixXd x)
            { 
                return (1.0 / (1.0 + 25.0 * (6*x.array()-1) * (6*x.array()-1))).matrix(); 
            };

// function f2
auto f2m = [](const Eigen::MatrixXd x)
            {
                return (1.0 + cos(M_PI*x.array())).matrix();
            };

////TASK E
// Computes the piecewise Lagrange interpolant
// and evaluates the interpolation error
double piecewiseLagrange_cheb(const int p, const int l)
{   
    double a = 0;
    double c = 1./3.;
    double b = 1;

    int N1 = std::pow(2,l);
    int N2 = 2*N1;
    
    // generate grid
    Eigen::VectorXd grid = Eigen::VectorXd::LinSpaced(N1+N2+1, a, b);
    
    // generate reference nodes
    Eigen::VectorXd knotsCheb = genKnotsCheb(p+1);
    Eigen::VectorXd knotsEqui = genKnotsEqui(p+1);
        
    // generate nodes
    Eigen::MatrixXd x(p+1, N1+N2);
    x.block(0, 0, p+1, N1) = genNodes(a, c, N1, knotsCheb);
    x.block(0, N1, p+1, N2) = genNodes(c, b, N2, knotsEqui);
    
    // generate data
    Eigen::MatrixXd y(p+1, N1+N2);
    y.block(0, 0, p+1, N1) = f1m(x.block(0, 0, p+1, N1));
    y.block(0, N1, p+1, N2) = f2m(x.block(0, N1, p+1, N2));
    
    // solve interpolant
    Newton ipolf(grid, x);
    ipolf.Interpolate(y);

    // compute interpolation error
    double error = ipolf.EvalError(f);
    
    return error;
}


double piecewiseLagrange_equi(const int p, const int l)
{   
    double a = 0;
    double c = 1./3.;
    double b = 1;

    int N1 = std::pow(2,l);
    int N2 = 2*N1;
    
    // generate grid
    Eigen::VectorXd grid = Eigen::VectorXd::LinSpaced(N1+N2+1, a, b);
    
    // generate reference nodes
    Eigen::VectorXd knotsEqui = genKnotsEqui(p+1);
        
    // generate nodes
    Eigen::MatrixXd x(p+1, N1+N2);
    x.block(0, 0, p+1, N1) = genNodes(a, c, N1, knotsEqui);
    x.block(0, N1, p+1, N2) = genNodes(c, b, N2, knotsEqui);
        
    // generate data
    Eigen::MatrixXd y(p+1, N1+N2);
    y.block(0, 0, p+1, N1) = f1m(x.block(0, 0, p+1, N1));
    y.block(0, N1, p+1, N2) = f2m(x.block(0, N1, p+1, N2));
    
    // solve interpolant
    Newton ipolf(grid, x);
    ipolf.Interpolate(y);

    // compute interpolation error
    double error = ipolf.EvalError(f);
    
    return error;
}
////END OF TASK E

////TASK F
// Run convergence test with respect to p
void convergenceWrtp() 
{
    int l=0;

    std::cout << "\n\nEquidistant nodes, convergence with respect to p:" << std::endl;
    for (int p=1; p<21; p++)
    {
        double error = piecewiseLagrange_equi(p, l);
        std::cout << p << "\t" << error << std::endl;
    }
    
    std::cout << "\n\nChebyshev nodes, convergence with respect to p:" << std::endl;
    for (int p=1; p<21; p++)
    {
        double error = piecewiseLagrange_cheb(p, l);
        std::cout << p << "\t" << error << std::endl;
    }
}
////END OF TASK F

// Run convergence test with respect to N
void convergenceWrtN() {
    
    int p=1;
    
    std::cout << "\n\nEquidistant nodes, convergence with respect to p:" << std::endl;
    for (int l=0; l<10; l++)
    {
        double error = piecewiseLagrange_equi(p, l);
        std::cout << l << "\t" << error << std::endl;
    }
    
    std::cout << "\n\nChebyshev nodes, convergence with respect to p:" << std::endl;
    for (int l=0; l<10; l++)
    {
        double error = piecewiseLagrange_cheb(p, l);
        std::cout << l << "\t" << error << std::endl;
    }
}


// A simple test
void test() {

    Eigen::VectorXd grid = Eigen::VectorXd::LinSpaced(2, -1, 1);
    Eigen::VectorXd knots = genKnotsEqui(5);
    Eigen::MatrixXd x = genNodes(-1, +1, 1, knots);
    Eigen::MatrixXd y = fm_test(x);
    
    Newton ipol(grid, x);
    ipol.Interpolate(y);
    std::cout << "\n\nCoefficients: " << ipol.GetCoeffs().transpose() << std::endl;
    // correct result: ipol.alpha = [0.0384615, 0.198939, 1.5252, -3.31565, 3.31565]   
}

// Computes the piecewise Lagrange interpolant 
// and evaluates the interpolation error
template<typename LambdaFn>
double piecewiseLagrange_test(LambdaFn F,
                         const Eigen::VectorXd &grid,
                         const Eigen::MatrixXd &x,
                         const Eigen::MatrixXd &y)
{
    // solve interpolant
    Newton ipolf(grid, x);
    ipolf.Interpolate(y);
    
    // compute interpolation error
    double error = ipolf.EvalError(F);
    
    return error;
}

// Run convergence test with respect to p
void test_convergenceWrtp() {
    
    double a = -1;
    double b = 1;
    int N = 1;
    Eigen::VectorXd grid = Eigen::VectorXd::LinSpaced(N+1, a, b);
    
    std::cout << "\n\nEquidistant nodes, convergence with respect to p:" << std::endl;
    for (int p=1; p<21; p++)
    {
        Eigen::VectorXd knots = genKnotsEqui(p+1);
        
        // generate nodes
        Eigen::MatrixXd x(p+1, N);
        x = genNodes(a, b, N, knots);
        
        // generate data
        Eigen::MatrixXd y(p+1, N);
        y = fm_test(x);
        
        double error = piecewiseLagrange_test(f_test, grid, x, y);
        std::cout << p << "\t" << error << std::endl;
    }
    
    std::cout << "\n\nChebyshev nodes, convergence with respect to p:" << std::endl;
    for (int p=1; p<21; p++)
    {
        Eigen::VectorXd knots = genKnotsCheb(p+1);
        
        // generate nodes
        Eigen::MatrixXd x(p+1, N);
        x = genNodes(a, b, N, knots);
        
        // generate data
        Eigen::MatrixXd y(p+1, N);
        y = fm_test(x);
        
        double error = piecewiseLagrange_test(f_test, grid, x, y);
        std::cout << p << "\t" << error << std::endl;
    }
}

// Run convergence test with respect to N
void test_convergenceWrtN() {
    
    double a = -1;
    double b = +1;
    int p = 1;
    
    std::cout << "\n\nEquidistant nodes, convergence with respect to N:" << std::endl;
    Eigen::VectorXd knotsEqui = genKnotsEqui(p+1);
    for (int l=0; l<10; l++)
    {
        int N = std::pow(2,l);
        Eigen::VectorXd grid = Eigen::VectorXd::LinSpaced(N+1, a, b);
    
        // generate nodes
        Eigen::MatrixXd x(knotsEqui.size(), N);
        x = genNodes(a, b, N, knotsEqui);
        
        // generate data
        Eigen::MatrixXd y(knotsEqui.size(), N);
        y = fm_test(x);
        
        double error = piecewiseLagrange_test(f_test, grid, x, y);
        std::cout << N << "\t" << error << std::endl;
    }
    
    std::cout << "\n\nChebyshev nodes, convergence with respect to N:" << std::endl;
    Eigen::VectorXd knotsCheb = genKnotsCheb(p+1);
    for (int l=0; l<10; l++)
    {
        int N = std::pow(2,l);
        Eigen::VectorXd grid = Eigen::VectorXd::LinSpaced(N+1, a, b);
    
        // generate nodes
        Eigen::MatrixXd x(knotsCheb.size(), N);
        x = genNodes(a, b, N, knotsCheb);
        
        // generate data
        Eigen::MatrixXd y(knotsCheb.size(), N);
        y = fm_test(x);
        
        double error = piecewiseLagrange_test(f_test, grid, x, y);
        std::cout << N << "\t" << error << std::endl;
    }
}


int main() {
	
    test();
    //test_convergenceWrtp();
    //test_convergenceWrtN();
    
    convergenceWrtp();
    //convergenceWrtN();

	return 0;
}


// End of file
